/** custom.css **/

.jupyter_cell.docutils.container {
    overflow-x: auto;
}

.pharmpy-snippet {
    margin: 1em 0;
}

.pharmpy-snippet div[class*="highlight-"] {
    margin: 0;
}

.pharmpy-snippet > .sphinx-tabs {
    margin: 0;
}

.pharmpy-snippet > .sphinx-tabs pre {
    border: 0;
    border-radius: 0;
    box-shadow: none;
}

/**
 * Could use upcoming :has(...) to use display:none fallback when not output is
 * produced. Instead of having to manually specify :hide-output:.
 */
.pharmpy-snippet.with-output > .jupyter_cell {
    padding: 1em;
    border: 1px solid #CCC;
    -moz-box-shadow: 2px 2px 4px rgba(87, 87, 87, 0.2);
    -webkit-box-shadow: 2px 2px 4px rgba(87, 87, 87, 0.2);
    box-shadow: 2px 2px 4px rgba(87, 87, 87, 0.2);
}

.pharmpy-snippet.with-output > .jupyter_cell > .cell_output > .output > div {
    overflow-x: auto;
}

table.dataframe {
    border-collapse: collapse;
    border: none;
    margin-bottom: 20px;
}

table.dataframe tr {
    border: none;
}

table.dataframe tr:nth-child(odd) {
	background: WhiteSmoke;
}

table.dataframe td, table.dataframe th {
    border: none;
    padding: 0.25em 0.75em 0.25em 0.75em;
    text-align: right;
}

table.dataframe thead, table.dataframe thead th {
	background: white;
    border-bottom: 1px solid black;
}
