"""
=======
Pharmpy
=======

Pharmpy is a python package for pharmacometric modeling.

Definitions
===========
"""

__version__ = '0.106.0'

import logging

logging.getLogger(__name__).addHandler(logging.NullHandler())
